from __future__ import annotations

from re import I
from telethon.sync import TelegramClient
from telethon.sessions import StringSession
import telethon
from telethon import utils
from telethon import types
import time
# import ipdb
from typing import List

#print("load " + '/'.join(__file__.split('/')[-2:]))

try:
    from .. import Os
    from .Database_src import SQLite
except:
    import sys 
    sys.path.append("..")
    import Os
    from Database_src import SQLite

class TelegramGeoAsync():
    def __init__(self):
        self.Long = None
        self.Lat = None 
        self.AccessHash = None
    
    def __repr__(self):
        return f"TelegramGeoAsync(Long={self.Long}, Lat={self.Lat}, AccessHash={self.AccessHash})"
        
    def __str__(self):
        return f"TelegramGeoAsync(Long={self.Long}, Lat={self.Lat}, AccessHash={self.AccessHash})"
        
class TelegramPhotoAsync():
    def __init__(self):
        self.ID = None
        self.AccessHash = 0
    
    def __repr__(self):
        return f"TelegramPhotoAsync(ID={self.ID}, AccessHash={self.AccessHash})"
        
    def __str__(self):
        return f"TelegramPhotoAsync(ID={self.ID}, AccessHash={self.AccessHash})"
        
# File and Audio
class TelegramFileAsync():
    def __init__(self):
        self.Name = ""
        self.Size = 0
        self.ID = None 
        self.AccessHash = 0
    
    def __repr__(self):
        return f"TelegramFileAsync(Name={self.Name}, Size={self.Size}, ID={self.ID}, AccessHash={self.AccessHash})"
        
    def __str__(self):
        return f"TelegramFileAsync(Name={self.Name}, Size={self.Size}, ID={self.ID}, AccessHash={self.AccessHash})"

class TelegramButtonAsync():
    def __init__(self, btn:telethon.tl.custom.messagebutton.MessageButton) -> None:
        self.btn = btn 

    async def Text(self) -> str:
        return self.btn.text
    
    async def Click(self):
        await self.btn.click()

    def __repr__(self):
        return f"TelegramButtonAsync(Text={self.btn.text})"
        
    def __str__(self):
        return f"TelegramButtonAsync(Text={self.btn.text})"

class TelegramMessageAsync():
    def __init__(self, client:TelegramClient):
        self.client = client
        self.peer:TelegramPeerAsync = None 
        self.message = None
        self.PeerType:str = None 
        self.Chat = TelegramPeerAsync(client=self.client)
        self.ID:int = None 
        self.Time:int = None 
        self.Action:str = None 
        self.File:TelegramFileAsync = None
        self.Photo:TelegramPhotoAsync = None
        self.Geo:TelegramGeoAsync = None
        self.Message:str = None
        self.User:TelegramPeerAsync = None
        self.Buttons:List[List[TelegramButtonAsync]] = None
        self.MessageRaw:str = None 
    
    async def DownloadMedia(self, SavePath:str=None) -> str | None:
        if not SavePath:
            if self.File:
                SavePath = self.File.Name
            elif self.Photo:
                SavePath = "photo.jpg"
        
        SavePath = Os.Path.Uniquify(SavePath)

        if not Os.Path.Exists(Os.Path.Basedir(SavePath)):
            Os.Mkdir(Os.Path.Basedir(SavePath))

        await self.client.download_media(self.message)
    
    async def Refresh(self) -> TelegramMessageAsync:
        return self.peer.Message(self.ID) 

    async def ClickButton(self, buttonText:str) -> bool:
        """
        If the button exists, click it and return True, otherwise return False
        
        :param buttonText: The text of the button you want to click
        :type buttonText: str
        :return: A boolean value.
        """
        if self.Buttons != None:
            for row in self.Buttons:
                for btn in row:
                    if btn.Text() == buttonText:
                        btn.Click()
                        return True 

        return False 
    
    def __repr__(self):
        return f"TelegramMessageAsync(PeerType={self.PeerType}, Chat={self.Chat}, ID={self.ID}, Time={self.Time}, Action={self.Action}, File={self.File}, Photo={self.Photo}, Message={self.Message}, User={self.User}, Button={self.Buttons})"
        
    def __str__(self):
        return f"TelegramMessageAsync(PeerType={self.PeerType}, Chat={self.Chat}, ID={self.ID}, Time={self.Time}, Action={self.Action}, File={self.File}, Photo={self.Photo}, Message={self.Message}, User={self.User}, Button={self.Buttons})"
        
class TelegramPeerAsync():
    def __init__(self, Type:str=None, Name:str=None, Username:str=None, ID:int=None, AccessHash:int=None, PhoneNumber:int=None, LangCode:str=None, client:TelegramClient=None):
        """
        :param Type: The type of the entity. Can be either "user" or "channel" (group)
        :type Type: str
        :param Name: The name of the user or channel
        :type Name: str
        :param Username: The username of the user or channel
        :type Username: str
        :param ID: The ID of the user or chat
        :type ID: int
        :param AccessHash: This is a unique identifier for a user or group. It is used to identify a user
        or group in a secure way
        :type AccessHash: int
        :param PhoneNumber: The phone number of the user
        :type PhoneNumber: int
        :param LangCode: The language code of the user
        :type LangCode: str
        """
        self.Type = Type # channel(group), user
        self.Name = Name # 名字, First Name + Last Name 或者 Title 
        self.Username = Username 
        self.ID = ID
        self.AccessHash = AccessHash
        self.PhoneNumber = PhoneNumber 
        self.LangCode = LangCode 
        self.Resolved = False # 是否已解析. 只设置一个ID, 解析之后就补上其它的字段.
        self.client = client # telethon.sync.TelegramClient
        self.entity = None 
    
    async def Message(self, id:str) -> TelegramMessageAsync:
        message = await self.client.get_messages(self.ID, ids=id)
        return await self.__wrapMsg(message)
    
    async def __wrapMsg(self, message) -> TelegramMessageAsync:
        # import ipdb
        # ipdb.set_trace()
        msg = TelegramMessageAsync(self.client)
        msg.peer = self
        msg.message = message
        msg.PeerType = self.Type 
        msg.Chat = self 
        msg.ID = message.id 
        msg.Time = int(message.date.timestamp())
        if message.action:
            msg.Action = message.action.to_dict()["_"]
        if message.media:
            if message.document:
                msg.File = TelegramFileAsync()
                msg.File.ID = message.document.id 
                msg.File.AccessHash = message.document.access_hash
                msg.File.Size = message.document.size 
                for attr in message.media.document.attributes:
                    if attr.to_dict()['_'] == "DocumentAttributeFilename":
                        msg.File.Name = attr.to_dict()['file_name']
            elif message.photo:
                msg.Photo = TelegramPhotoAsync()
                msg.Photo.ID = message.photo.id
                msg.Photo.AccessHash = message.photo.access_hash
            elif message.geo:
                msg.Geo = TelegramGeoAsync()
                msg.Geo.AccessHash = message.geo.access_hash
                msg.Geo.Lat = message.geo.lat
                msg.Geo.Long = message.geo.long
            # else: 
            #     import ipdb 
            #     ipdb.set_trace()
            #     print(message)
        if message.message:
            msg.Message = message.message
        if message.text:
            msg.MessageRaw = message.text
        if message.from_id:
            msg.User = TelegramPeerAsync(ID=message.from_id.user_id, client=self.client)
        
        # ipdb.set_trace()

        if message.buttons != None:
            buttons = []
            for row in message.buttons:
                btns = []
                for btn in row:
                    btns.append(TelegramButtonAsync(btn))
                buttons.append(btns)
            
            msg.Buttons = buttons

        return msg
    
    async def Messages(self, limit:int=100, offset:int=0) -> list[TelegramMessageAsync]:
        """
        It takes a chat object, and returns a list of messages in that chat.
        
        :param limit: The maximum number of messages to be returned, defaults to 100
        :type limit: int (optional)
        :param offset: The offset of the first message to be returned, defaults to 0
        :type offset: int (optional)
        :return: A list of TelegramMessageAsync objects
        """
        res = []
        getmessage = await self.client.get_messages(self.ID, limit=limit, offset_id=offset)
        for message in getmessage:
            msg = await self.__wrapMsg(message)
            res.append(msg)
        return res

    async def Resolve(self) -> TelegramPeerAsync:
        """
        Resolve Peer, get information by peer id. 
        """
        if self.ID:
            self.entity = await self.client.get_entity(self.ID)
            # import ipdb
            # ipdb.set_trace()
            if type(self.entity) == telethon.tl.types.Channel:
                self.Type = "channel"
                self.Name = self.entity.title
            elif type(self.entity) == telethon.tl.types.User:
                self.Type = "user"
                self.Name = " ".join([i for i in filter(lambda x: x != None, [self.entity.first_name, self.entity.last_name])])
                self.PhoneNumber = self.entity.phone
                self.LangCode = self.entity.lang_code

            self.AccessHash = self.entity.access_hash
            self.Username = self.entity.username 
            self.ID = self.entity.id
        
        return self

    def __repr__(self):
        if self.Type == "user":
            return f"TelegramPeerAsync(Type={self.Type}, Name={self.Name}, Username={self.Username}, ID={self.ID}, AccessHash={self.AccessHash}, PhoneNumber={self.PhoneNumber})"
        else:
            return f"TelegramPeerAsync(Type={self.Type}, Name={self.Name}, Username={self.Username}, ID={self.ID}, AccessHash={self.AccessHash})"

    def __str__(self):
        return self.__repr__()
    
    async def SendMessage(self, message:str):
        if not self.entity:
            self.Resolve()

        await self.client.send_message(self.entity, message)

# It's a wrapper for the `telethon` library that allows you to use it in a more Pythonic way
class TelegramAsync():
    def __init__(self, appid:str, apphash:str, sessionfile:str, phone:str=None):
        self.client = TelegramClient(sessionfile, appid, apphash, device_model="Samsung S22 Ultra", system_version="Android 10.0.0", app_version="4.0.2") 
        # self.client = TelegramClient(StringSession(sessionString), appid, apphash)
        if phone:
            self.client.start(phone=phone)
        else:
            self.client.start()

        # me = self.client.get_me()
        # print(me.stringify())
        self.sessionfile = sessionfile + ".session"

    async def SessionString(self) -> str:
        """
        It takes the session object from the client object and saves it to a string
        :return: The session string is being returned.
        """
        return self.client.session.save()
    
    async def PeerByUsername(self, username:str) -> TelegramPeerAsync:
        """
        根据Username解析一个Peer, 有速率限制
        
        :param username: The username of the user/channel you want to send the message to
        :type username: str
        """
        tp = TelegramPeerAsync()

        tp.client = self.client

        try:
            obj = await self.client.get_entity(username)
        except (ValueError, TypeError):
            time.sleep(1)
            obj = await self.client.get_entity(username)
        if type(obj) == telethon.tl.types.Channel:
            tp.Type = "channel"
            tp.Name = obj.title
        elif type(obj) == telethon.tl.types.User:
            tp.Type = "user"
            tp.Name = " ".join(filter(lambda x: x != None, [obj.first_name, obj.last_name]))
        
        tp.AccessHash = obj.access_hash
        tp.Username = obj.username 
        tp.ID = obj.id

        self.client.session.save()

        return tp
    
    async def PeerByIDAndHash(self, ID:int, Hash:int, Type:str="channel") -> TelegramPeerAsync:
        """
        根据ID和Hash返回一个Peer, 没有速率限制
        
        :param ID: The ID of the user or group
        :type ID: int
        :param Hash: The hash value of the peer, which can be obtained by calling the GetPeerHash method
        of the TelegramPeerAsync object
        :type Hash: int
        :param Type: The type of the peer, which can be "channel", "group", or "user", defaults to
        channel
        :type Type: str (optional)
        :return: TelegramPeerAsync
        """
        if Type in ["channel", "group"]:
            tp = types.PeerChannel(ID)
        elif Type == "user":
            tp = types.PeerUser(ID)
        else:
            raise Exception(f"未知的类型:{Type}")
        peerid = utils.get_peer_id(tp)

        try:
            return await self.PeerByUsername(peerid)
        except (ValueError, TypeError):
            self.client.session.save() # save all data to sqlite database session file to avoide database lock
            db = SQLite(self.sessionfile)
            (db.Table("entities").
                Data({
                    "id": peerid, 
                    "hash": Hash,
                }).Insert())
            db.Close()
            try:
                peer = await self.PeerByUsername(peerid)
            except (ValueError, TypeError):
                time.sleep(1)
                peer = await self.PeerByUsername(peerid)
            self.client.session.save() # save the entity we just resolved to the database

            return peer 

    async def GetMe(self) -> TelegramPeerAsync:
        me = await self.client.get_me()
        p = await self.PeerByIDAndHash(ID=me.id, Hash=me.access_hash, Type="user")
        return await p.Resolve()

if __name__ == "__main__":
    pass






from urllib.parse import urlparse, quote_plus, unquote

import requests
from urllib.parse import urlparse, urljoin

#print("load " + '/'.join(__file__.split('/')[-2:]))

class URLParseResult():
    def __init__(self, Schema:str, Host:str, Port:int, User:str, Pass:str, Path:str, Query:str, Fragment:str):
        self.Schema = Schema
        self.Host = Host    
        self.Port = Port    
        self.User = User    
        self.Pass = Pass    
        self.Path = Path    
        self.Query = Query   
        self.Fragment = Fragment
    
    def __repr__(self):
        return f"URLParseResult(Schema={self.Schema}, Host={self.Host}, Port={self.Port}, User={self.User}, Pass={self.Pass}, Path={self.Path}, Query={self.Query}, Fragment={self.Fragment})"

    def __str__(self):
        return f"URLParseResult(Schema={self.Schema}, Host={self.Host}, Port={self.Port}, User={self.User}, Pass={self.Pass}, Path={self.Path}, Query={self.Query}, Fragment={self.Fragment})"

class URL():
    def __init__(self, url:str):
        self.url = url 
    
    def Parse(self) -> URLParseResult:
        """
        It parses the URL and returns the URLParseResult object.
        :return: A URLParseResult object.
        """
        # 有时候是 example.com/abc?key=value, 这样没法解析, 所以加一个头才能解析
        if not self.url.startswith("http://") and not self.url.startswith("https://"):
            url = 'http://' + self.url 
        else:
            url = self.url 

        res = urlparse(url)

        # ipdb.set_trace()

        # 如果是自己加的协议头, 就不写到schema里面了
        if url == self.url:
            Schema = res.scheme
        else:
            Schema = None
        Path = res.path
        Query = res.query 
        Fragment = res.fragment

        if '@' not in res.netloc:
            User = None
            Pass = None
        else:
            u = res.netloc.split("@")[0]
            if ':' in u:
                User = u.split(":")[0]
                Pass = u.split(":")[1]
            elif len(u) != 0:
                User = u 
                Pass = None
            else:
                User = None 
                Pass = None
        # print(res)
        h = res.netloc 
        if '@' in res.netloc:
            h = res.netloc.split("@")[1]
        
        if ':' not in h:
            Host = h 
            if Schema == "http":
                Port = 80
            elif Schema == "https":
                Port = 443 
            else:
                Port = None
        else:
            Host = h.split(":")[0]
            Port = int(h.split(":")[1])
        
        return URLParseResult(Schema, Host, Port, User, Pass, Path, Query, Fragment)
    
    def Encode(self) -> str:
        return quote_plus(self.url)
    
    def Decode(self) -> str:
        return unquote(self.url)

    def GetRedirectChain(self) -> list[str]:
        """
        传入一个URL，返回所有301或302跳转的链接（包括最终的跳转目标）。如果传入的url不可访问则返回的list为空.

        参数:
        url (str): 初始URL。

        返回:
        list: 包含所有跳转链接的列表，按跳转顺序排列。
        """
        url = self.url

        redirect_chain = []
        session = requests.Session()
        while True:
            # print("trying:", url)
            try:
                response = session.get(url, allow_redirects=False, stream=True, timeout=15)
                response.close()  # 立即关闭连接，避免读取body
            except Exception as e:
                # print(f"Error fetching {url}: {e}")
                break

            # 获取完整的URI
            current_uri = urlparse(response.url)
            current_uri = f"{current_uri.scheme}://{current_uri.netloc}{current_uri.path}"
            # print('got:', current_uri)
            if current_uri not in redirect_chain:
                redirect_chain.append(current_uri)
            else:
                break

            if response.status_code in (301, 302):
                # 获取跳转目标URL
                location = response.headers.get('Location')
                if location:
                    # 确保跳转URL是绝对路径
                    url = urljoin(response.url, location)
                else:
                    break
            else:
                break

        return redirect_chain

if __name__ == "__main__":
    # u = URL("http://user:pass@docs.python.org:8897/3/library/urllib.parse.html?highlight=params&k=v#url-parsing")
    # print(u.Parse())

    # u = URL("example.com?title=правовая+защита")
    # print(u.Encode())

    # u = URL(u.Encode())
    # print(u.Decode())

    # u = URL("example.com/abc?key=value")
    # print(u.Parse())

    # u = URL("ss://YWVzLTI1Ni1jZmI6YW1hem9uc2tyMDU@54.95.169.40:443#%e6%97%a5%e6%9c%ac%3dtg%e9%a2%91%e9%81%93%3a%40bpjzx2%3d1")
    # print(u.Parse())

    print(URL('chrome-extension://aapbdbdomjkkjkaonfhkkikfgjllcleb/popup_css_compiled.css').Parse())


import listparser
import requests

#print("load " + '/'.join(__file__.split('/')[-2:]))

class rssFeed():
    def __init__(self, title:str, url:str):
        self.Title = title 
        self.URL = url
    
    def __str__(self) -> str:
        return f"RSSFeed(Title={self.Title} URl={self.URL})"

    def __repr__(self) -> str:
        return self.__str__()

def Opml(opmlurl:str) -> list[rssFeed]:
    res = []
    for i in listparser.parse(requests.get(opmlurl).content)['feeds']:
        res.append(rssFeed(i["title"], i["url"]))
    
    return res

if __name__ == "__main__":
    for i in Opml("https://wechat2rss.xlab.app/opml/sec.opml"):
        print(i)


# -*- coding: utf-8 -*-

from __future__ import division

import math
from .base import BasePaginator
from ..support.collection import Collection
from ..utils import deprecated


class LengthAwarePaginator(BasePaginator):
    def __init__(self, items, total, per_page, current_page=None, options=None):
        """
        Constructor

        :param items: The items being paginated
        :type items: mixed

        :param total: Total number of results
        :type total: int

        :param per_page: The number of results per page
        :type per_page: int

        :param current_page: The current page of results
        :type current_page: int

        :param options: Extra options to set
        :type options: dict
        """
        if options is not None:
            for key, value in options.items():
                setattr(self, key, value)

        self.total = total
        self.per_page = per_page
        self.last_page = int(math.ceil(total / per_page))
        self.current_page = self._set_current_page(current_page, self.last_page)
        if isinstance(items, Collection):
            self._items = items
        else:
            self._items = Collection.make(items)

    def _set_current_page(self, current_page, last_page):
        """
        Get the current page for the request.

        :param current_page: The current page of results
        :type current_page: int

        :param last_page: The last page of results
        :type last_page: int

        :rtype: int
        """
        if not current_page:
            current_page = self.resolve_current_page()

        if current_page > last_page:
            if last_page > 0:
                return last_page

            return 1

        if not self._is_valid_page_number(current_page):
            return 1

        return current_page

    def has_more_pages(self):
        """
        Determine if there are more items in the data source.

        :rtype: int
        """
        return self.current_page < self.last_page

    @deprecated
    def to_dict(self):
        """
        Alias for serialize.

        :rtype: list
        """
        return self.serialize()

    def serialize(self):
        """
        Convert the object into something JSON serializable.

        :rtype: list
        """
        return self._items.serialize()

    def to_json(self, **options):
        return self._items.to_json(**options)



# -*- coding: utf-8 -*-

from orator.migrations import Migrator, DatabaseMigrationRepository
from .base_command import BaseCommand


class RollbackCommand(BaseCommand):
    """
    Rollback the last database migration.

    migrate:rollback
        {--d|database= : The database connection to use.}
        {--p|path= : The path of migrations files to be executed.}
        {--P|pretend : Dump the SQL queries that would be run.}
        {--f|force : Force the operation to run.}
    """

    def handle(self):
        """
        Executes the command.
        """
        if not self.confirm_to_proceed(
            "<question>Are you sure you want to rollback the last migration?:</question> "
        ):
            return

        database = self.option("database")
        repository = DatabaseMigrationRepository(self.resolver, "migrations")

        migrator = Migrator(repository, self.resolver)

        self._prepare_database(migrator, database)

        pretend = self.option("pretend")

        path = self.option("path")

        if path is None:
            path = self._get_migration_path()

        migrator.rollback(path, pretend)

        for note in migrator.get_notes():
            self.line(note)

    def _prepare_database(self, migrator, database):
        migrator.set_connection(database)



from __future__ import annotations

import numpy as np
import cv2
import types
import typing
import time
import os
import copy
import flask

from ..Thread import Thread
from .. import Time
from .Ratelimit_src import RateLimit
from .. import Socket
from ..String import String
from .. import Lg
from .. import Base64
from .. import Http
from .. import Json
from .. import Random

#print("load " + '/'.join(__file__.split('/')[-2:]))

here = os.path.dirname(os.path.abspath(__file__))

netssd = None 
classesssd = ["background", "aeroplane", "bicycle", "bird", "boat",
    "bottle", "bus", "car", "cat", "chair", "cow", "diningtable",
    "dog", "horse", "motorbike", "person", "pottedplant", "sheep",
    "sofa", "train", "tvmonitor"]
colorsssd = np.random.uniform(0, 255, size=(len(classesssd), 3))

netyolo = None 
classesyolo = []
colorsyolo = []

colorsapiserver = {}

class cvStreamFrameObjectDetectionResult():
    def __init__(self, detections, objectDetectModel, frame) -> None:
        self.detections = detections
        self.objectDetectModel = objectDetectModel
        self.frame = frame 

        if self.objectDetectModel == "YOLO":
            self.prepareYOLO()

    def drawBySSD(self, frame:cvStreamFrame, filterAbove:int=0, filterName:list=[]) -> cvStreamFrame:
        """
        > Draws a rectangle around each object detected by the SSD model, and displays the object's name
        and confidence level
        
        :param frame: the frame to draw on
        :type frame: cvStreamFrame
        :param filterAbove: Filter out detections with confidence below this value, defaults to 0. 百分比, 100为完全匹配.
        :type filterAbove: int (optional)
        :param filterName: list of strings, names of objects to filter out
        :type filterName: list
        :return: A frame with the detections drawn on it.
        """
        frame = copy.deepcopy(frame)
        (H, W) = frame.frame.shape[:2]

        for i in np.arange(0, self.detections.shape[2]):
            # extract the confidence (i.e., probability) associated
            # with the prediction
            confidence = self.detections[0, 0, i, 2]
            # filter out weak detections by ensuring the `confidence`
            # is greater than the minimum confidence
            if confidence > filterAbove / 100:
                # extract the index of the class label from the
                # detections list
                idx = int(self.detections[0, 0, i, 1])
                name = classesssd[idx]
                if name in filterName:
                    continue 
                # if the class label is not a car, ignore it
                # if classesssd[idx] != "car":
                #     continue
                # compute the (x, y)-coordinates of the bounding box
                # for the object
                box = self.detections[0, 0, i, 3:7] * np.array([W, H, W, H])
                (startX, startY, endX, endY) = box.astype("int")
                
                cv2.rectangle(
                    frame.frame, 
                    (startX, startY), 
                    (endX, endY), 
                    (0, 255, 0), 1)

                cv2.putText(frame.frame, '%s: %.2f%%' % (name, confidence*100),
                        (startX+25, startY+30),
                        cv2.FONT_HERSHEY_DUPLEX,
                        1, (0, 255, 0), 1) 
        
        return frame
    
    def prepareYOLO(self):
        self.class_ids = []
        self.confidences = []
        self.boxes = []
        conf_threshold = 0.5
        nms_threshold = 0.4

        Width = self.frame.shape[1]
        Height = self.frame.shape[0]

        for out in self.detections:
            for detection in out:
                scores = detection[5:]
                class_id = np.argmax(scores)
                confidence = scores[class_id]

                center_x = int(detection[0] * Width)
                center_y = int(detection[1] * Height)
                w = int(detection[2] * Width)
                h = int(detection[3] * Height)
                x = center_x - w / 2
                y = center_y - h / 2
                self.class_ids.append(class_id)
                self.confidences.append(float(confidence))
                self.boxes.append([x, y, w, h])
        
        self.indices = cv2.dnn.NMSBoxes(self.boxes, self.confidences, conf_threshold, nms_threshold)

    def drawByYOLO(self, frame:cvStreamFrame, filterAbove:int=0, filterName:list=[]) -> cvStreamFrame:
        frame = copy.deepcopy(frame)

        for i in self.indices:
            try:
                box = self.boxes[i]
            except:
                i = i[0]
                box = self.boxes[i]
            
            x = box[0]
            y = box[1]
            w = box[2]
            h = box[3]

            label = str(classesyolo[self.class_ids[i]])
            color = colorsyolo[self.class_ids[i]]
            confidence = self.confidences[i]
            cv2.rectangle(frame.frame, (round(x), round(y)), (round(x+w), round(y+h)), color, 1)
            # print('%s: %.2f%%' % (label, confidence*100))
            # print(x+25, y+30)
            cv2.putText(frame.frame, '%s: %.2f%%' % (label, confidence*100),
                        (round(x)+5, round(y)+15),
                        cv2.FONT_HERSHEY_DUPLEX,
                        0.5, color, 0.5) 
    
        return frame 

    def drawByAPIServer(self, frame:cvStreamFrame, filterAbove:int=0, filterName:list=[]) -> cvStreamFrame:
        frame = copy.deepcopy(frame)

        global colorsapiserver

        for i in self.detections:
            if i["name"] not in colorsapiserver:
                colorsapiserver[i["name"]] = (Random.Int(0, 256), Random.Int(0, 256), Random.Int(0, 256))

            cv2.rectangle(
                frame.frame, 
                (i['coordinate'][0][0], i['coordinate'][0][1]), 
                (i['coordinate'][1][0], i['coordinate'][1][1]), 
                (0, 255, 0), 1)

            cv2.putText(frame.frame, '%s: %.2f%%' % (i['name'], i['confidence']*100),
                    (i['coordinate'][0][0]+25, i['coordinate'][0][1]+30),
                    cv2.FONT_HERSHEY_DUPLEX,
                    1, (0, 255, 0), 1) 
        
        return frame

    def Draw(self, frame:cvStreamFrame, filterAbove:int=0, filterName:list=[]) -> cvStreamFrame:
        if self.objectDetectModel == "SSD":
            return self.drawBySSD(frame, filterAbove, filterName)
        elif self.objectDetectModel == "YOLO":
            return self.drawByYOLO(frame, filterAbove, filterName)  
        elif self.objectDetectModel.startswith("APIServer"):
            return self.drawByAPIServer(frame, filterAbove, filterName)  

    def Objects(self) -> dict:
        resp = {}

        if self.objectDetectModel == "SSD":
            for i in np.arange(0, self.detections.shape[2]):
                confidence = self.detections[0, 0, i, 2]
                idx = int(self.detections[0, 0, i, 1])
                name = classesssd[idx]
                resp[name] = confidence
        elif self.objectDetectModel == "YOLO":
            for i in self.indices:
                name = str(classesyolo[self.class_ids[i]])
                confidence = self.confidences[i]
                resp[name] = confidence
        elif self.objectDetectModel.startswith("APIServer"):
             for i in self.detections:
                resp[i['name']] = i['confidence']
        else:
            raise Exception("需要先加载模型: SetSSDModelForObjectDetect or SetYoloModelForObjectDetect or SetAPIServerForObjectDetect")
        
        return resp

class cvStreamFrameDifference():
    def __init__(self, cnts) -> None:
        self.cnts = cnts

    def Draw(self, frame:cvStreamFrame) -> cvStreamFrame:
        frame = copy.deepcopy(frame)
        for c in self.cnts:
            (x, y, w, h) = cv2.boundingRect(c)
            cv2.rectangle(frame.frame, (x, y), (x + w, y + h), (0, 255, 0), 1)
        
        return frame

    def HasDifference(self) -> bool:
        return len(self.cnts) != 0

class cvStreamFrame():
    def __init__(self, frame, objectDetectModel) -> None:
        self.frame = frame
        self.grayFrame = None 
        self.objectDetectModel = objectDetectModel
        self.createTime = Time.Now()

    def __grayFrame(self):
        # print(type(self.grayFrame))
        if type(self.grayFrame) == types.NoneType:
            gray = cv2.cvtColor(self.frame, cv2.COLOR_BGR2GRAY)
            self.grayFrame = cv2.GaussianBlur(gray, (21, 21), 0)
        
        return self.grayFrame
    
    def objectsBySSD(self) -> cvStreamFrameObjectDetectionResult:
        # convert the frame to a blob and pass the blob through the
        # network and obtain the detections
        blob = cv2.dnn.blobFromImage(self.frame, size=(600, 600), ddepth=cv2.CV_8U)
        netssd.setInput(blob, scalefactor=1.0/127.5, mean=[127.5, 127.5, 127.5])
        detections = netssd.forward()

        return cvStreamFrameObjectDetectionResult(detections, self.objectDetectModel, self.frame)

    def objectsByYOLO(self) -> cvStreamFrameObjectDetectionResult:
        scale = 0.00392

        blob = cv2.dnn.blobFromImage(self.frame, scale, (416,416), (0,0,0), True, crop=False)

        netyolo.setInput(blob)

        layer_names = netyolo.getLayerNames()
        try:
            output_layers = [layer_names[i - 1] for i in netyolo.getUnconnectedOutLayers()]
        except:
            output_layers = [layer_names[i[0] - 1] for i in netyolo.getUnconnectedOutLayers()]


        detections = netyolo.forward(output_layers)

        return cvStreamFrameObjectDetectionResult(detections, self.objectDetectModel, self.frame)

    def objectsByAPIServer(self) -> cvStreamFrameObjectDetectionResult:
        (flag, encodedImage) = cv2.imencode(".jpg", self.frame)
        encodedImage = bytearray(encodedImage)
        encodedImage = Base64.Encode(encodedImage)

        a = self.objectDetectModel.split("|")
        server = a[1]
        model = a[2]

        if not server.startswith("http://") and not server.startswith("https://"):
            server = "https://" + server 

        while True:
            try:
                resp = Http.PostJson(server + "/object-detect", {
                    "model": model,
                    "data": encodedImage,
                }, timeout=60)
                break
            except Exception as e:
                Lg.Warn("调用API服务器识别失败:", e)
                Time.Sleep(5)
                pass 

            if resp.StatusCode > 500:
                Lg.Warn("调用API服务器识别失败: " + str(resp.StatusCode))
                Time.Sleep(1)
                Lg.Trace("重试")
            elif resp.StatusCode == 500:
                raise Exception("调用API服务器识别失败: " + str(resp.StatusCode))
            else:
                break 
        
        res = Json.Loads(resp.Content)
        if res['code'] != 200:
            raise Exception("调用API服务器识别失败:" + res['message'])

        return cvStreamFrameObjectDetectionResult(res['result'], self.objectDetectModel, self.frame)

    def Objects(self, objectDetectModel:str=None) -> cvStreamFrameObjectDetectionResult:
        if objectDetectModel == None:
            if self.objectDetectModel == "SSD":
                return self.objectsBySSD()
            elif self.objectDetectModel == "YOLO":
                return self.objectsByYOLO()
            elif self.objectDetectModel.startswith("APIServer"):
                return self.objectsByAPIServer()
            else:
                raise Exception("需要先加载模型: SetSSDModelForObjectDetect or SetYoloModelForObjectDetect or SetAPIServerForObjectDetect")
        
        else:
            if objectDetectModel == "SSD":
                return self.objectsBySSD()
            elif objectDetectModel == "YOLO":
                return self.objectsByYOLO()
            elif objectDetectModel.startswith("APIServer"):
                return self.objectsByAPIServer()
            else:
                raise Exception("需要先加载模型: SetSSDModelForObjectDetect or SetYoloModelForObjectDetect or SetAPIServerForObjectDetect")
        

    def Compare(self, frame:cvStreamFrame, size:int=250) -> cvStreamFrameDifference:
        # Compare the difference between current frame and the background frame 
        frameDelta = cv2.absdiff(self.__grayFrame(), frame.__grayFrame())
        thresh = cv2.threshold(frameDelta, 25, 255, cv2.THRESH_BINARY)[1]

        # Expand the threshold image to fill the hole, and then find the contour on the threshold image
        thresh = cv2.dilate(thresh, None, iterations=2)
        (cnts, _) = cv2.findContours(thresh.copy(), cv2.RETR_EXTERNAL,
            cv2.CHAIN_APPROX_SIMPLE)
        
        ccnts = []
        for c in cnts:
            if cv2.contourArea(c) < size:
                continue
            ccnts.append(c)
        
        return cvStreamFrameDifference(cnts)

    def Show(self, title:str="", wait:bool=False):
        cv2.imshow(title, self.frame) 
        if wait == True:
            cv2.waitKey(0)
            cv2.destroyAllWindows()

    def Save(self, path:str):
        try:
            cv2.imwrite(path, self.frame)
        except Exception as e:
            if not os.path.exists(os.path.dirname(path)):
                os.makedirs(os.path.dirname(path))
                cv2.imwrite(path, self.frame)
            else:
                raise e

    def Resize(self, precent:float) -> cvStreamFrame:
        """
        转换图片/帧的大小, precent为百分比. 大于100为放大, 小于100为缩小. 
        
        :param precent: The percentage of the original size you want to resize the image to
        :type precent: int
        :return: A cvStreamFrame object
        """
        precent = precent / 100
        frame = cv2.resize(self.frame, (0, 0), fx=precent, fy=precent)
        return cvStreamFrame(frame, self.objectDetectModel)

    def Bright(self, times:int=None) -> cvStreamFrame:
        """
        > It takes an image and a number, and returns a brighter version of the image
        
        :param times: How many times to brighter
        :type times: int
        :return: A cvStreamFrame object.
        """
        # print("times:", times)
        if times == None:
            if self.Brightness() != "dark":
                return self 
            
            brightcount = 2
            while True:
                # print(brightcount)
                frame = copy.deepcopy(self)
                frame = frame.Bright(brightcount)
                if frame.Brightness() != "dark":
                    return frame
                
                brightcount += 1
        else:
            if times == 1:
                return self
            
            img2 = cv2.add(self.frame, self.frame)
            if times == 2:
                return cvStreamFrame(img2, self.objectDetectModel) 
            else:
                for _ in range(2, times):
                    # print("add", _)
                    img2 = cv2.add(img2, self.frame)

                    # cvStreamFrame(img2).Show(wait=True)
            
                return cvStreamFrame(img2, self.objectDetectModel)
    
    def BrightCheck(self) -> int:
        if self.Brightness() != "dark":
            return 1
            
        brightcount = 2
        while True:
            # print(brightcount)
            frame = copy.deepcopy(self)
            frame = frame.Bright(brightcount)
            if frame.Brightness() != "dark":
                return brightcount
            
            brightcount += 1
        
    def Brightness(self) -> str:
        # 把图片转换为单通道的灰度图
        gray_img = cv2.cvtColor(self.frame, cv2.COLOR_BGR2GRAY)
        
        # 获取形状以及长宽
        img_shape = gray_img.shape
        height, width = img_shape[0], img_shape[1]
        size = gray_img.size
        # 灰度图的直方图
        hist = cv2.calcHist([gray_img], [0], None, [256], [0, 256])
        
        # 计算灰度图像素点偏离均值(128)程序
        ma = 0
        reduce_matrix = np.full((height, width), 128)
        shift_value = gray_img - reduce_matrix
        shift_sum = sum(map(sum, shift_value))

        da = shift_sum / size
        # 计算偏离128的平均偏差
        for i in range(256):
            ma += (abs(i-128-da) * hist[i])
        m = abs(ma / size)
        # 亮度系数
        k = abs(da) / m
        # print(k, da)
        if k[0] > 1:
            # 过亮
            if da > 0:
                # print("过亮")
                return 'light'
            else:
                # print("过暗")
                return 'dark'
        else:
            # print("亮度正常")
            return 'normal'

    def Rotate(self, side:int) -> cvStreamFrame:
        """
        旋转或者镜像

        :param side: 0, 1, -1分别为逆时针旋转90度, 180度, 左右镜像
        :type side: int
        :return: A cvStreamFrame object
        """
        frame = cv2.flip(cv2.transpose(self.frame), side)
        return cvStreamFrame(frame, self.objectDetectModel)

    def Text(self, text:str, x:int=10, y:int=-10) -> cvStreamFrame:
        if x < 0:
            x = self.frame.shape[1] + x
        if y < 0:
            y = self.frame.shape[0] + y
        cv2.putText(self.frame, text, (x, y), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (0, 0, 255), 1)
        return self
    
    def Size(self) -> typing.Tuple[int, int]:
        """
        宽和高, w and h
        """
        return self.frame.shape[0], self.frame.shape[1]

class StreamSync():
    def __init__(self, source:int|str) -> None:
        """
        source可以是数字, 0, 1, 2, 表示摄像头的编号. 可以是本地视频文件的路径. 可以是远程摄像头的http地址.
        
        :param source: The source of the video. 
        :type source: int|str
        """
        self.source = source
        self.stream = cv2.VideoCapture(source)
    
        self.objectDetectModel = None 

        self.fpss = []
        self.frameCountSec = 0
        self.lastTimeForCaclFPS = None

        self.FPS = 0
        self.FPSAverage = 0

    def SetSSDModelForObjectDetect(self, prototxt:str="MobileNetSSD_deploy.prototxt.txt", caffemodel:str="MobileNetSSD_deploy.caffemodel"):
        self.objectDetectModel = "SSD"
        global netssd

        netssd = cv2.dnn.readNetFromCaffe(
            prototxt, 
            caffemodel,
        )

    def SetYoloModelForObjectDetect(self, weights:str="yolov4.weights", config:str="yolov4.cfg", classes:str="yolov4.txt"):
        self.objectDetectModel = "YOLO"
        global classesyolo
        global colorsyolo
        global netyolo

        if netyolo == None:
            with open(classes, 'r') as f:
                classesyolo = [line.strip() for line in f.readlines()]

            colorsyolo = np.random.uniform(0, 255, size=(len(classesyolo), 3))

            netyolo = cv2.dnn.readNet(weights, config)
    
    def SetAPIServerForObjectDetect(self, server:str, model:str="yolo"):
        self.objectDetectModel = f"APIServer|{server}|{model}"

    def Close(self):
        try:
            self.stream.release()
        except:
            pass 
        try:
            cv2.destroyAllWindows()
        except:
            pass
    
    def Get(self) -> cvStreamFrame:
        (grabbed, frame) = self.stream.read()

        if not grabbed:
            return 
        
        self.frameCountSec += 1
        
        if self.frameCountSec != 0 and (self.lastTimeForCaclFPS == None or Time.Now() - self.lastTimeForCaclFPS >= 1):
            self.FPS = self.frameCountSec

            self.fpss.append(self.frameCountSec)

            while len(self.fpss) > 180:
                self.fpss.pop(0)

            self.FPSAverage = int(sum(self.fpss)/len(self.fpss))
            
            self.lastTimeForCaclFPS = Time.Now()
            self.frameCountSec = 0
        
        return cvStreamFrame(frame, self.objectDetectModel)

    def __iter__(self) -> typing.Iterator[cvStreamFrame]:
        while True:
            (grabbed, frame) = self.stream.read()

            if not grabbed:
                self.Close()
                return 
            
            self.frameCountSec += 1
        
            if self.frameCountSec != 0 and (self.lastTimeForCaclFPS == None or Time.Now() - self.lastTimeForCaclFPS >= 1):
                self.FPS = self.frameCountSec

                self.fpss.append(self.frameCountSec)

                while len(self.fpss) > 180:
                    self.fpss.pop(0)

                self.FPSAverage = int(sum(self.fpss)/len(self.fpss))
                
                self.lastTimeForCaclFPS = Time.Now()
                self.frameCountSec = 0

            key = cv2.waitKey(1) & 0xFF
            if key == ord("q"):
                break

            yield cvStreamFrame(frame, self.objectDetectModel)
            
    def __enter__(self):
        return self 
    
    def __exit__(self, exc_type, exc_value, traceback):
        try:
            self.Close()
        except:
            pass

class StreamAsync():
    def __init__(self, source:int|str) -> None:
        """
        source可以是数字, 0, 1, 2, 表示摄像头的编号. 可以是远程摄像头的http地址.
        
        :param source: The source of the video. 
        :type source: int|str
        """
        self.source = source

        self.closed = False

        self.lastFrameUpdateTime = None
        self.lastFrame = None 

        self.lastGetFrameTime = None 

        self.lastTimeForCaclFPS = None
        self.frameCountSec = 0
        self.fpss = []

        self.FPS = None
        self.FPSAverage = None
        self.webVideoFeedLastGet = {} 

        self.FPSRead = None 

        self.objectDetectModel = "SSD"

        Thread(self.run)

        # print(0)
        while self.lastFrameUpdateTime == None:
            Time.Sleep(0.1)

        # print(2)
        while self.FPS == None:
            self.Get()
    
    def SetSSDModelForObjectDetect(self, prototxt:str="MobileNetSSD_deploy.prototxt.txt", caffemodel:str="MobileNetSSD_deploy.caffemodel"):
        self.objectDetectModel = "SSD"
        global netssd

        netssd = cv2.dnn.readNetFromCaffe(
            prototxt, 
            caffemodel,
        )

    def SetYoloModelForObjectDetect(self, weights:str="yolov4.weights", config:str="yolov4.cfg", classes:str="yolov4.txt"):
        self.objectDetectModel = "YOLO"
        global classesyolo
        global colorsyolo
        global netyolo

        if netyolo == None:
            with open(classes, 'r') as f:
                classesyolo = [line.strip() for line in f.readlines()]

            colorsyolo = np.random.uniform(0, 255, size=(len(classesyolo), 3))

            netyolo = cv2.dnn.readNet(weights, config)
    
    def SetAPIServerForObjectDetect(self, server:str, model:str="yolo"):
        self.objectDetectModel = f"APIServer|{server}|{model}"
        
    def Close(self, destroyAllWindows:bool=True):
        self.closed = True
        try:
            if self.stype == "c":
                Lg.Trace("关闭stream")
                self.stream.release()
            elif self.stype == "n":
                Lg.Trace("关闭socket")
                self.stream.Close()
        except Exception as e:
            Lg.Trace("关闭有异常:", e)
            pass 
    
        if destroyAllWindows:
            try:
                Lg.Trace("销毁所有窗口")
                cv2.destroyAllWindows()
            except Exception as e:
                Lg.Trace("销毁有异常:", e)
                pass
        Lg.Trace("关闭完成")

    def openStream(self):
        if type(self.source) == str and self.source.startswith("socket://"):
            Lg.Trace("打开网络接口")
            self.stype = "n"
            _ = String(self.source).RegexFind("socket://(.+?):(.+)")
            # print(_[0])
            self.stream = Socket.TCP.Connect(_[0][1], int(_[0][2])).PacketConnection()
        else:
            self.stream = cv2.VideoCapture(self.source)
            self.stype = "c"
        Lg.Trace("开启stream完成, 模式:", self.stype)

    def run(self):
        self.openStream()

        frameCountSec = 0
        lastSec = Time.Now()
        while True:
            if self.stype == "c":
                (grabbed, frame) = self.stream.read()

                if not grabbed:
                    Lg.Trace("没有抓到帧")
                    try:
                        Lg.Trace("关闭stream")
                        self.stream.release()
                    except Exception as e:
                        Lg.Trace("关闭有异常:", e)
                        pass 
                    time.sleep(1)
                    Lg.Trace("重新尝试")
                    self.openStream()
                    time.sleep(1)
                    continue 
                
            elif self.stype == "n":
                # print(2)
                try:
                    self.stream.Send("d")
                    imgdata = self.stream.Recv()
                except Exception as e:
                    Lg.Trace("从socket抓帧报错了:", e)
                    try:
                        Lg.Trace("关闭socket")
                        self.stream.Close()
                    except Exception as e:
                        Lg.Trace("关闭有异常:", e)
                        pass 
                    time.sleep(1)
                    Lg.Trace("重新尝试")
                    self.openStream()
                    time.sleep(1)
                    continue 
                
                nparr = np.fromstring(imgdata, np.uint8)
                frame = cv2.imdecode(nparr, cv2.IMREAD_COLOR)
                # print(3)
            
            frameCountSec += 1
            
            if Time.Now() - lastSec > 1 and frameCountSec != 0:
                self.FPSRead = frameCountSec
                frameCountSec = 0
                lastSec = Time.Now()

            # # print(4)
            self.lastFrame = cvStreamFrame(frame, self.objectDetectModel)
            self.lastFrameUpdateTime = Time.Now()
        
    def socketserverrunner(self, tc:Socket.TCP.PacketConnection):
        lastgettime = None
        while True:
            if self.closed == True:
                break 

            # print(504)
            # print(505, tc.Recv())
            # print(506)

            while lastgettime == self.lastFrameUpdateTime:
                if self.FPSRead != None:
                    Time.Sleep(1/self.FPSRead)
                else:
                    Time.Sleep(1/30)

            (flag, encodedImage) = cv2.imencode(".jpg", self.lastFrame.frame)
            if not flag:
                if self.FPSRead != None:
                    Time.Sleep(1/self.FPSRead)
                else:
                    Time.Sleep(1/30)
                continue

            encodedImage = encodedImage.tobytes()
            try:
                tc.Send(encodedImage)
            except:
                break

            lastgettime = self.lastFrameUpdateTime

        tc.Close()

    def socketServer(self, ipaddr:str, port:int):
        for tc in Socket.TCP.Listen(ipaddr, port):
            tc = tc.PacketConnection()
            Thread(self.socketserverrunner, tc)
    
    def SocketServer(self, ipaddr:str="0.0.0.0", port:int=7283):
        Thread(self.socketServer, ipaddr, port)

    def Get(self) -> cvStreamFrame:
        while True:
            #print(self.frameCountSec)
            if self.frameCountSec != 0 and (self.lastTimeForCaclFPS == None or Time.Now() - self.lastTimeForCaclFPS >= 1):
                self.FPS = self.frameCountSec

                self.fpss.append(self.frameCountSec)

                while len(self.fpss) > 180:
                    self.fpss.pop(0)

                self.FPSAverage = int(sum(self.fpss)/len(self.fpss))
                
                self.lastTimeForCaclFPS = Time.Now()
                self.frameCountSec = 0

            if self.lastGetFrameTime != self.lastFrameUpdateTime:
                self.frameCountSec += 1
                #print(self.lastGetFrameTime, self.lastFrameUpdateTime)
                self.lastGetFrameTime = self.lastFrameUpdateTime
                return self.lastFrame

            if self.closed == True:
                return None 
            
            if self.FPS != None:
                Time.Sleep(1/self.FPS)
            else:
                Time.Sleep(1/30)
    
    def webResponseImageGenerate(self) -> bytes:
        while True:
            if self.closed == True:
                break 
            
            (flag, encodedImage) = cv2.imencode(".jpg", self.lastFrame.frame)
            # ensure the frame was successfully encoded
            if not flag:
                if self.FPSRead != None:
                    Time.Sleep(1/self.FPSRead)
                else:
                    Time.Sleep(1/30)
                continue
            
            yield(b'--frame\r\n' b'Content-Type: image/jpeg\r\n\r\n' + bytearray(encodedImage) + b'\r\n')

            if self.FPSRead != None:
                Time.Sleep(1/self.FPSRead)
            else:
                Time.Sleep(1/30)

    def webResponseVideoFeed(self):
        return flask.Response(self.webResponseImageGenerate(),
            mimetype = "multipart/x-mixed-replace; boundary=frame")
    
    def RunWebServer(self, ipaddr:str="0.0.0.0", port:int=9987, url:str="/camera/video"):
        flaskapp = flask.Flask("whatevername")
        flaskapp.add_url_rule(url, 'webResponseVideoFeed', self.webResponseVideoFeed)

        Thread(flaskapp.run, ipaddr, port)

    def __iter__(self) -> typing.Iterator[cvStreamFrame]:
        while True:
            frame = self.Get()

            if frame == None:
                return
            
            key = cv2.waitKey(1) & 0xFF
            if key == ord("q"):
                self.closed = True
                break

            yield frame
            
    def __enter__(self):
        return self 
    
    def __exit__(self, exc_type, exc_value, traceback):
        try:
            self.Close()
        except:
            pass

def LoadImage(path:str, objectDetectModel:str="SSD", weights:str="yolov4.weights", config:str="yolov4.cfg", classes:str="yolov4.txt") -> cvStreamFrame:
    if objectDetectModel == "YOLO":
        global classesyolo
        global colorsyolo
        global netyolo

        with open(classes, 'r') as f:
            classesyolo = [line.strip() for line in f.readlines()]

        colorsyolo = np.random.uniform(0, 255, size=(len(classesyolo), 3))

        netyolo = cv2.dnn.readNet(weights, config)

    return cvStreamFrame(cv2.imread(path), objectDetectModel)

class VideoWriter():
    def __init__(self, path:str, fps:int, width:int, height:int) -> None:
        fourcc = cv2.VideoWriter_fourcc('m', 'p', '4', 'v')
        self.writer = cv2.VideoWriter(path, fourcc, fps, (height, width))
        self.closed = False

    def Write(self, frame:cvStreamFrame):
        if self.closed == False:
            self.writer.write(frame.frame) 

    def Close(self):
        if self.closed == False:
            self.closed = True 
            self.writer.release() 

class ComputerVision:
    VideoWriter
    LoadImage
    StreamAsync
    StreamSync

if __name__ == "__main__":
    import os
    import datetime 
    
    # web camera
    # stream = Stream("http://10.129.129.207:8080/video")

    # usb camera
    # stream = Stream(0)

    # print(stream.FPS())

    # Video file
    # stream = StreamSync(os.getenv("HOME") + "/Desktop/1080p/221105.mp4")

    # stream.RunWebServer()

    # for frame in stream:
    #     frame.Text(str(stream.FPS) + "/" + str(stream.FPSAverage)).Show("")

    # bg = stream.Get()
    # for frame in stream:
    #     frame.Compare(bg).Draw(frame).Text(datetime.datetime.now().strftime("%A %d %B %Y %I:%M:%S%p")).Show("test")

    # bg = stream.Get().Bright(5).Rotate(0)
    # for frame in stream:
    #     frame = frame.Bright(5).Rotate(0)
    #     frame.Compare(bg).Draw(frame).Text(datetime.datetime.now().strftime("%A %d %B %Y %I:%M:%S%p")).Show("test")

    # for frame in stream:
    #     frame.Objects().Draw(frame, filterAbove=70).Show("")

    # frame = stream.Get()
    # w, h = frame.Size()
    # print(w,h )
    # writer = VideoWriter("video.mp4", 25, w, h)

    # for _ in range(0, 250):
    #     writer.Write(stream.Get())
    
    # writer.Close()

    #####################3

    # stream = Tools.ComputerVision.StreamSync(0)
    # stream.SetAPIServerForObjectDetect("example.com")
    # for frame in stream:
    #     frame = frame.Rotate(0).Text(f"fps:{stream.FPS}")
    #     frame = frame.Objects().Draw(frame)
    #     frame.Show()
    pass 


# -*- coding: utf-8 -*-

from ..utils import decode
from ..utils import PY2
from .connection import Connection
from ..query.grammars.mysql_grammar import MySQLQueryGrammar
from ..query.processors.mysql_processor import MySQLQueryProcessor
from ..schema.grammars import MySQLSchemaGrammar
from ..schema import MySQLSchemaBuilder
from ..dbal.mysql_schema_manager import MySQLSchemaManager


class MySQLConnection(Connection):

    name = "mysql"

    def get_default_query_grammar(self):
        return MySQLQueryGrammar(marker=self._marker)

    def get_default_post_processor(self):
        return MySQLQueryProcessor()

    def get_schema_builder(self):
        """
        Return the underlying schema builder.

        :rtype: orator.schema.SchemaBuilder
        """
        if not self._schema_grammar:
            self.use_default_schema_grammar()

        return MySQLSchemaBuilder(self)

    def get_default_schema_grammar(self):
        return self.with_table_prefix(MySQLSchemaGrammar(self))

    def get_schema_manager(self):
        return MySQLSchemaManager(self)

    def begin_transaction(self):
        self._reconnect_if_missing_connection()

        try:
            self._connection.autocommit(False)
        except Exception as e:
            if self._caused_by_lost_connection(e):
                self.reconnect()
                self._connection.autocommit(False)
            else:
                raise

        super(MySQLConnection, self).begin_transaction()

    def commit(self):
        if self._transactions == 1:
            self._connection.commit()
            self._connection.autocommit(True)

        self._transactions -= 1

    def rollback(self):
        if self._transactions == 1:
            self._transactions = 0

            self._connection.rollback()
            self._connection.autocommit(True)
        else:
            self._transactions -= 1

    def _get_cursor_query(self, query, bindings):
        if not hasattr(self._cursor, "_last_executed") or self._pretending:
            return super(MySQLConnection, self)._get_cursor_query(query, bindings)

        if PY2:
            return decode(self._cursor._last_executed)

        return self._cursor._last_executed



import csv 

#print("load " + '/'.join(__file__.split('/')[-2:]))

class Reader():
    def __init__(self, fpath:str, withHeader:bool=True):
        self.fpath = fpath 
        self.fd = open(self.fpath)
        self.csvrd = csv.reader(self.fd, delimiter=',', quotechar='"', escapechar='\\')
        self.withHeader = withHeader
        if self.withHeader:
            self.headers = next(self.csvrd)
    
    def Read(self) -> dict | list:
        while True:
            try:
                r = next(self.csvrd)
                break 
            except StopIteration:
                raise StopIteration
            except Exception:
                pass 
        
        if self.withHeader:
            row = {}
            for idx in range(len(self.headers)):
                try:
                    row[self.headers[idx]] = r[idx]
                except IndexError:
                    row[self.headers[idx]] = "" 
        else:
            return r 
        
        return row
    
    def __iter__(self):
        while True:
            try:
                yield self.Read()
            except StopIteration:
                self.Close()
                return 
    
    def Close(self):
        self.fd.close()
        
class Writer():
    def __init__(self, fpath:str, mode:str="w", autoflush:bool=True):
        self.fpath = fpath
        self.fd = open(self.fpath, mode, newline='')
        self.csvwd = csv.writer(self.fd, delimiter=',', quotechar='"', escapechar='\\', doublequote=False)# , quoting=csv.QUOTE_NONE)
        self.fdmode = mode
        self.headers = None
        if self.fdmode != "w":
            try:
                self.headers = Reader(fpath).headers
            except StopIteration:
                self.fdmode = "w"
        self.autoflush = autoflush

    def SetHeaders(self, *headers):
        self.headers = headers
        if self.fdmode == "w":
            self.csvwd.writerow(headers)
    
    def Write(self, row:dict[str]):
        r = []
        for header in self.headers:
            if header in row:
                r.append(row[header])
            else:
                r.append("")
        
        self.csvwd.writerow(r)
        if self.autoflush:
            self.fd.flush()

    def Close(self):
        self.fd.close()
    
    def Flush(self):
        self.fd.flush()
    
    def __enter__(self):
        return self 
    
    def __exit__(self, exc_type, exc_value, traceback):
        try:
            self.Close()
        except:
            pass

if __name__ == "__main__":
    w = Writer("test.csv")

    w.SetHeaders("h1", "h2")

    w.Write({"h1": "v1", "h2": '"v2,kkk|'})
    w.Write({"h1": "v,1", "h2": '"v222'})
    w.Write({"h1": "3", "h2": '"99kkk'})

    w.Close()

    # test.csv
    # h1,h2
    # v1,"\"v2,kkk|"
    # "v,1",\"v222
    # 3,\"99kkk

    r = Reader("test.csv")
    print(r.Read()) # {'h1': 'v1', 'h2': '"v2,kkk|'}

    for row in r:
        print(row) 
        # {'h1': 'v,1', 'h2': '"v222'}
        # {'h1': '3', 'h2': '"99kkk'}
    
    w = Writer("test.csv", "a")
    w.Write({"h1": "4", "h2": '5'}) 
    w.Write({"h1": "6", "h3": '7'}) # 6,

    w = Writer("test1.csv", "a")
    w.SetHeaders("h1", "h2")
    w.Write({"h1": "4", "h2": '5'}) 
    w.Write({"h1": "6", "h3": '7'}) # 6,
